#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "cablinux.h"
#include "cabstruct.h"
#include "cabheader.h"
#include "cabfolder.h"
#include "cabfile.h"
#include "cabdata.h"
#include "cabwrite.h"

// global variables ----------------------------------------------------------
char *files[1000];
char *outfile;
FILE *fpout;


/* ** count bytes in file ** */
int sizefile(char *filename)
{
	int size=0;
	FILE *fp=fopen(filename,"rb");
	if(!fp) 
	{ 
		printf("Error: could not open %s\n", filename); 
		exit(1); 
	}
	while(fgetc(fp)!=EOF) 
	{ 
		++size; 
	}
	fclose(fp);
	return size;
}

/* ** calculate the number of datablocks we will need:
      cabinet files are written in blocks of 32768 bytes ** */
int number_of_datablocks(int nof)
{
	int i, size=0;
	for(i=0;i<nof;++i)
	{ 
		size+=sizefile(*(files+i)); 
	}
	return size/DATABLOCKSIZE+1 ;
}

/* ** put all data into a temporary file ** */
int maketempfile(int nof, FILE *fptemp)
{
 	int i;
  	FILE *fp;
  	char *tmp;

  	for(i=0;i<nof;++i)
    	{
      		fp=fopen(files[i],"rb");
      		if(!fp) 
		{ 
			printf("Error: could not open %s\n", files[i]); 
			return 0; 
		}
      		tmp = (byte *) calloc(sizefile(files[i]),sizeof(byte));
      		fread(tmp,sizefile(files[i]), 1, fp);
      		fwrite((byte *) tmp, sizefile(files[i]), 1, fptemp);
      		free(tmp);
      		fclose(fp);
    	}

  	return 1;
}

/* ** strip path from filename ** */
char *make_wop(char *filename)
{
	char *tmp;
	if(!(tmp=strrchr(filename,'/'))) tmp=filename;
	else *tmp++;
	return tmp;
}


//----------------------------------------------------------------------------
// main
//----------------------------------------------------------------------------
int main(int argc, char *argv[])
{
	struct cabheader ch;
	struct cabfolder cfo;
	struct cabfile cfi[100], tmpcfi;
	struct cabdata cd[MAXDATABLOCKS];

	int nof=0;				/* number of files */
	int nod=0;				/* number of datablocks */
	int i,offsetdata=0,size1=0,j;
	int makewop=0; 				/* create wop files */
	FILE *fptemp;
	char *fptempfilename;
	long int pos=0;

	float kb;

// with or without paths ? 
	if((argc>1) && !strcmp(argv[1],"--wop")) makewop=1;

	if(argc<3+makewop) 
	{ 
		printf("usage: cablinux [--wop] infiles outfile.cab\n"); 
		printf("       --wop    create files without pathnames\n");
		return 0; 
	}

	printf("Cablinux v1.0.3 (2001) by Rien (rien@filii.luc.ac.be)\n");

// list files to add 
	nof=argc-2-makewop; 
	for(i=0;i<nof+makewop;++i) 
	{ 
		files[i]=argv[1+i+makewop]; 
	}

	nod=number_of_datablocks(nof);
	outfile=argv[nof+1+makewop]; 

/* ** HEADER ** */
	ch = cabheader_init(ch,"MSCF",1,nof,0,1234,0);
	ch = cabheader_offsetfiles(ch,CFI_start);

/* ** FOLDER ** */
	cfo = cabfolder_init(cfo,nod,0);

/* ** DATABLOCK ** */
	for(j=0;j<nof;++j) { size1 += sizefile(files[j]); }
	if(size1 < DATABLOCKSIZE)
	{
		cd[0] = cabdata_init(cd[0],0);
		if(nof==1)
		{
			cd[0] = cabdata_ncbytes(cd[0],sizefile(files[0]));
			cd[0] = cabdata_nubytes(cd[0],sizefile(files[0]));
		}
		else
		{
			cd[0] = cabdata_ncbytes(cd[0],size1);
			cd[0] = cabdata_nubytes(cd[0],size1);
		}
	}
	else
	{
		for(i=0;i<nod;++i)
		{
			if(i!=nod-1)
			{
				cd[i] = cabdata_ncbytes(cd[i],DATABLOCKSIZE);
				cd[i] = cabdata_nubytes(cd[i],DATABLOCKSIZE); 	
			}
			else
			{
				size1-=DATABLOCKSIZE*(nod-1);
				cd[i] = cabdata_ncbytes(cd[i],size1);
				cd[i] = cabdata_nubytes(cd[i],size1); 	
			}
		}
	}

/* ** HEADER size ** */
	size1=CFI_start+nof*16; // 16 bytes for first part of cabfile-structure 
	for(i=0;i<nof;++i) 
	{ 
		size1+=strlen(makewop?make_wop(files[i]):files[i]) +1; // +1 for \0 !!!
	} 
	size1+=nod*8; // 8 bytes for first part of cabdata-structure
	for(i=0;i<nod;++i) 
	{ 
		size1+=cd[i].nubytes; 
	} 
	ch = cabheader_size(ch, size1); 
	kb = ch.size/1024;
	printf("total size: %ld bytes (%3.2f Kb)\n", ch.size,kb);
	
/* ** FILE ** */
	for(i=0;i<nof;++i)
	{
		cfi[i] = cabfile_init(cfi[i],sizefile(files[i]),0,makewop?make_wop(files[i]):files[i]);
		// offset 1st cabfile-entry=0, 2nd is offset 1st + filesize 1st, ...
		if(i==0) cabfile_uoffset(cfi[i],0);
		else 
		{
			tmpcfi = cfi[i-1];
			cfi[i] = cabfile_uoffset(cfi[i],tmpcfi.uoffset+sizefile(files[i-1]));
		}
	}
	offsetdata=0;
	for(i=0;i<nof;++i) 
	{ 
		offsetdata+=16+strlen(makewop?make_wop(files[i]):files[i]) +1; 
	}
	cfo = cabfolder_offsetdata(cfo,CFI_start+offsetdata);

/* ** WRITE ** */
// set temp filename
	fptempfilename = NULL;
	fptempfilename = tmpnam(fptempfilename);
// fill temp file
	if(!(fptemp = fopen(fptempfilename,"wb")))
	{ printf("Error: could not open temp file.\n"); 
	  exit(1); }
	if(!maketempfile(nof,fptemp)) return 0;
	fclose(fptemp);

// open output file
	if(!(fpout=fopen(outfile,"wb"))) 
	{ 	printf("Error outfile: could not open %s\n", outfile); 
		exit(1); 
	}

// write cabheader
	if( !cabheaderwrite(ch,fpout) )
	{ 
		printf("Error writing cabheader\n"); 
		exit(1); 
	}
// write cabfolder
	if( !cabfolderwrite(cfo,fpout) )  
	{ 
		printf("Error writing cabfolder\n"); 
		exit(1); 
	}
// write cabfile
	printf("adding files: ");
	for(i=0;i<nof;++i) 
	{
	    	if( !cabfilewrite(cfi[i],fpout) )  
		{ 
			printf("Error writing cabfile\n"); 
			exit(1); 
		} 
		else printf(".");
	}
// open temp file for reading
	if(!(fptemp = fopen(fptempfilename,"rb")))
	{ printf("Error: could not open temp file %s.\n", fptempfilename); 
	  exit(1); }
// write cabdata
	for(i=0;i<nod;++i) 
	{
		pos = cabdatawrite(cd[i],fpout,pos,fptemp);
		if(pos==-1) 
		{ 
			printf("error writing cdata\n"); 
			exit(1); 
		}
	} 

//close output file
	fclose(fpout);
// close temp file
	fclose(fptemp);
// delete temp file
	remove(fptempfilename);

	printf(" done.\n");

	exit(0);
}
